@@ -16,7 +16,7 @@ along with notmuch.  If not, see <http://www.gnu.org/licenses/>.
 
 Copyright 2010 Sebastian Spaeth <Sebastian@SSpaeth.de>
 """
-import sys
+
 from ctypes import CDLL, Structure, POINTER
 
 #-----------------------------------------------------------------------------
@@ -26,38 +26,7 @@ try:
 except:
     raise ImportError("Could not find shared 'notmuch' library.")
 
-
-if sys.version_info[0] == 2:
-    class Python3StringMixIn(object):
-        def __str__(self):
-            return unicode(self).encode('utf-8')
-
-
-    def _str(value):
-        """Ensure a nicely utf-8 encoded string to pass to libnotmuch
-
-        C++ code expects strings to be well formatted and
-        unicode strings to have no null bytes."""
-        if not isinstance(value, basestring):
-            raise TypeError("Expected str or unicode, got %s" % type(value))
-        if isinstance(value, unicode):
-            return value.encode('UTF-8')
-        return value
-else:
-    class Python3StringMixIn(object):
-        def __str__(self):
-            return self.__unicode__()
-
-
-    def _str(value):
-        """Ensure a nicely utf-8 encoded string to pass to libnotmuch
-
-        C++ code expects strings to be well formatted and
-        unicode strings to have no null bytes."""
-        if not isinstance(value, str):
-            raise TypeError("Expected str, got %s" % type(value))
-        return value.encode('UTF-8')
-
+from .compat import Python3StringMixIn, encode_utf8 as _str
 
 class Enum(object):
     """Provides ENUMS as "code=Enum(['a','b','c'])" where code.a=0 etc..."""
